//
//  HttpUtils.swift
//  CepmApp
//
//  Created by 楚海 on 2020/12/25.
//

import Foundation
import Alamofire
import os

public class HttpUtils{
    
    /// Json报文头
    public static let jsonHeaders: HTTPHeaders = [
        .contentType("application/json;charset=utf-8")
    ]
    
    /// 日志对象
    private static let log = Logger()
    
    
    /// 执行Http请求
    /// - Parameters:
    ///   - url: 请求地址
    ///   - params: 请求参数
    ///   - method: 请求方式
    ///   - headers: 报文头
    ///   - complation: 请求完成后的处理方法
    public class func doRequest(_ url: URLConvertible,params: Parameters? = nil,method: HTTPMethod,headers: HTTPHeaders=HttpUtils.jsonHeaders) -> Request{
        return AF.request(url,
                          method:  method,
                          parameters:  params,
                          encoding: JSONEncoding.default,
                          headers: headers)
    }
    
    /// 执行Http请求
    /// - Parameters:
    ///   - url: 请求地址
    ///   - params: 请求参数
    ///   - method: 请求方式
    ///   - headers: 报文头
    ///   - complation: 请求完成后的处理方法
    ///   - successHandler: 成功后处理函数
    ///   - errorHandler: 失败后的处理函数
    public class func doRequest(_ url: String,params: Parameters? = nil,method: HTTPMethod,headers: HTTPHeaders=HttpUtils.jsonHeaders,successHandler:@escaping (_ data:Any) ->Void,errorHandler:@escaping (_ error:AFError)-> Void){
        AF.request(url,
                   method:  method,
                   parameters:  params,
                   encoding: JSONEncoding.default,
                   headers: headers).responseJSON{response in
                    switch response.result{
                    case .success(let value):
                        log.debug("http \(method.rawValue) 请求成功; [url = \(url)]")
                        successHandler(value)
                    case .failure(let err):
                        log.warning("Http \(method.rawValue) 请求失败;[url = \(url);Error = \(err.errorDescription!)]")
                        errorHandler(err)
                    }
                   }
    }
    
    
    /// 执行Http请求
    /// - Parameters:
    ///   - url: 请求地址
    ///   - params: 请求参数
    ///   - method: 请求方式
    ///   - headers: 报文头
    ///   - complation: 请求完成后的处理方法
    public class func doRequest<T:BaseEntry>(_ url: String,params: Parameters? = nil,method: HTTPMethod,headers: HTTPHeaders=HttpUtils.jsonHeaders, complation:@escaping ( _ entry:T?)->Void){
        AF.request(url,
                   method:  method,
                   parameters:  params,
                   encoding: JSONEncoding.default,
                   headers: headers).responseObject { (response : AFDataResponse<T>) in
                    switch response.result{
                    case .success(let entry):
                        log.debug("http \(method.rawValue) 请求成功; [url = \(url)]")
                        complation(entry)
                    case .failure(let error):
                        log.warning("Http \(method.rawValue) 请求失败;[url = \(url);Error = \(error.errorDescription!)]")
                        print(error)
                    }
                   }
    }
    
    /// 执行Post请求
    /// - Parameters:
    ///   - url: 请求地址
    ///   - params: 请求参数
    ///   - headers: 报文头
    ///   - complation: 请求完成后的处理方法
    public class func doPost<T:BaseEntry>(_ url: String,params: Parameters,headers: HTTPHeaders=HttpUtils.jsonHeaders, complation:@escaping (_ entry:T?)->Void){
        HttpUtils.doRequest(url,
                            params:  params,
                            method:  .post,
                            headers: headers,
                            complation:complation)
    }
    
    /// 执行GET请求
    /// - Parameters:
    ///   - url: 请求地址
    ///   - params: 请求参数
    ///   - headers: 报文头
    ///   - complation: 请求完成后的处理方法
    public class func doGet<T:BaseEntry>(_ url: String,params: Parameters,headers: HTTPHeaders=HttpUtils.jsonHeaders, complation:@escaping (_ entry:T?)->Void){
        HttpUtils.doRequest(url,
                            params:  params,
                            method:  .get,
                            headers: headers,
                            complation:complation)
    }
    
    public class func doPostSyn<T:BaseEntry>(_ url: String,params: Parameters,headers: HTTPHeaders=HttpUtils.jsonHeaders)->T?{
        let response:AFDataResponse<T> =   AF.request(url,
                                                      method:  .post,
                                                      parameters:  params,
                                                      encoding: JSONEncoding.default,
                                                      headers: headers).responseObjectSyn()
        switch response.result {
        case .success(let entry):
            log.debug("http POST 请求成功; [url = \(url)] [JSON = \(entry.toJSON())]")
            return entry
        case .failure(let error):
            log.warning("Http POST 请求失败;[url = \(url);Error = \(error.errorDescription!)]")
            return nil
        }
    }
    
    /// 文件异步上传
    /// - Parameters:
    ///   - url: 请求路径
    ///   - filePath: 文件地址(不包含文件名)
    ///   - onSuccess: 成功回调函数
    ///   - onError: 失败回调函数
    public class func doUploadFile<T:BaseEntry>(_ url: String, filePath: String,fileName: String, onSuccess:@escaping (_ data:T?) ->Void,onError:@escaping (_ error:AFError)-> Void) {
        let fileURL:URL = URL.init(fileURLWithPath: filePath + fileName)
        let data:Data = try! Data(contentsOf: fileURL)
        AF.upload(multipartFormData: { (MFData) in
            MFData.append(data, withName: "file", fileName: fileName)
        }, to: url).responseObject { (response : AFDataResponse<T>) in
            switch response.result{
            case .success(let entry):
                log.debug("http upload 请求成功; [url = \(url)]; [Response = \(entry.toJSON())]")
                onSuccess(entry)
            case .failure(let error):
                log.warning("Http upload 请求失败;[url = \(url);Error = \(error.errorDescription!)]")
                print(error)
                onError(error)
            }
        }
    }
    
    
    /// 文件上传同步方法
    /// - Parameters:
    ///   - url: 请求地址
    ///   - filePath: 文件地址(不包含文件名)
    ///   - fileName: 文件名称(后缀名小写)
    /// - Returns: T:BaseEntry 实体
    public class func doUploadFileSync<T:BaseEntry>(_ url: String, filePath: String, fileName: String) -> T? {
        let fileURL:URL = URL.init(fileURLWithPath: filePath + fileName)
        let data:Data = try! Data(contentsOf: fileURL)
        let upLoadRequest:UploadRequest =  AF.upload(multipartFormData: { (MFData) in
            MFData.append(data, withName: "file", fileName: fileName)
        }, to: url)
        let response:AFDataResponse<T> = upLoadRequest.validate().responseObjectSyn()
        switch response.result {
        case .success(let entry):
            log.debug("http upload 请求成功; [url = \(url)]; [Response = \(entry.toJSON())]")
            return entry
        case .failure(let error):
            log.warning("Http upload 请求失败;[url = \(url);Error = \(error.errorDescription!)]")
            return nil
        }
        
        
    }
}
